import sys, types, os, os.path, fnmatch, re, Prophet




from Keywords import Keyword as Kw, Set as KwS
from Directories import Set as DS
from Prophet.Categories import *
from Prophet import msg




class Containers(DS) :
	"""Set of valid directories which contain .desktop files"""

	def isValid(self, container) :
		# A valid container must contain at least one .desktop file in its directory
		# Note that these dirs may have the subdirs which should be visited as well
		for x in os.walk(container) :
			if len(fnmatch.filter(x[2], "*.desktop")) :
				return True
		return False




def scan() :
	"""Scan through all containers and return a list of found valid entries"""
	result = []
	msg("  desktop...", newline = False)
	for c in containers :
		for w in os.walk(c) :
			for f in fnmatch.filter(w[2], "*.desktop") :
				try :
					result.append(Application(os.path.join(w[0], f)))
				except (NotDesktop, Prophet.NotFound) :
					pass
	msg(" %d apps found" % len(result))
	return result




def fromPrefixes() :
	"""Extract the containers from all known prefixes"""
	c = []
	for x in Prophet.prefixes :
		for z in ("share/applications", "share/applnk", "share/gnome/apps") :
			c.append(os.path.join(x, z))
	return Containers(*c)




containers = fromPrefixes() + Containers("/usr/share/applications", "~/.local/share/applications")




_section	= re.compile("\s*?\[\s*(.*)\s*\].*?") # Match section definition, e.g. [Global Settings]
_skip		= re.compile("\s*[\#\;].*") # Match comments - lines starting with # or ;
_entry		= re.compile("\s*?(.*)\s*?\=\s*?(.*)\s*?") # Match any valid entry, e.g. key = value
_pkey		= re.compile("(\w+)\s*\[\s*(\w+)\s*\]") # Match parametrized key, e.g. key[param]




_DE = Kw("Desktop Entry")




def _parse(desktop) :
	section = []
	inSection = False
	try :
		xs = open(desktop).readlines()
	except IOError :
		raise NotDesktop("i/o error while reading .desktop file")
	for x in xs :
		if re.match(_skip, x) :
			continue
		rx = re.match(_section, x)
		if rx :
			inSection = (Kw(rx.group(1)) == _DE)
			continue
		if inSection :
			rx = re.match(_entry, x)
			if rx :
				section.append( (rx.group(1), rx.group(2)) ) # (key, value)
	if len(section) :
		return _parseEntries(section)
	raise NotDesktop("[Desktop Entry] section is either absent or empty")





def _parseEntries(entries) :
	e = {}
	for key, val in entries :
		rx = re.match(_pkey, key)
		if rx :
			pass # TODO : _parseLocale
		else :
			e[key] = _parseValue(val)
	return e




def _parseValue(value) :
	vl = [x.strip() for x in value.split(";") if not (x.strip() == "")]
	if len(vl) > 1 :
		return tuple(vl)
	else :
		return value.strip()




class NotDesktop(Exception) :
	"""Exception is raised when parser fails to decode entry"""




def _string(x) :
	"""Pick up the most appropriate string representation out of all specified variants
		Choice is made with respect to current locale and type of specified argument"""
	if type(x) == types.DictType :
		return x[""]
	# FIXME : locale handling
	else :
		# Now argument is assumed to be a string
		return x




def _bool(x) :
	"""Convert string to boolean"""
	if x.lower() in ("yes", "true", "1") :
		return True
	else :
		return False




class Application(Prophet.Application) :

	def __init__(self, desktop) :
		x = _parse(desktop)

		try :
			if not (x["Type"] == "Application") :
				raise NotDesktop("Type is not Application")
		except KeyError :
			raise NotDesktop("Type entry not found")

		try :
			cmd = x["Exec"]
			# Exec entry may contain the %-prefixed arguments that make no sense for us
			# and should be filtered out. If they're there, set exename to anything prior first
			# argument encountered
			scmd = cmd.split()
			if len(scmd) < 2 :
				self.exename = cmd
			else :
				self.exename = scmd[0].strip()
				self.exeargs = ""
				for i in range(1, len(scmd)) :
					if "%" in scmd[i] :
						break # First %arg encountered terminates the list
					try :
						if scmd[i].startswith("-") and "%" in scmd[i+1] :
							break # Filter out things like `-caption "%c"'
					except IndexError :
						pass
					self.exeargs += scmd[i] + " "
				self.exeargs = self.exeargs.rstrip()
				if not len(self.exeargs) :
					del self.exeargs
		except KeyError :
			try :
				# As a last resort, check for TryExec entry. It should not contain arguments so consume it all.
				self.exename = x["TryExec"]
			except KeyError :
				raise NotDesktop("Exec/TryExec entry is absent or invalid")
				# If both entries are absent, this desktop file is totally useless for us

		for n in ("Name", "GenericName") :
			self.name = _string(x[n])
			break

		try :
			self.comment = _string(x["Comment"])
		except KeyError :
			pass

		try :
			self.keywords = KwS(*x["Categories"])
		except KeyError :
			# Absence of the categories renders this desktop entry worthless since there's
			# no clue on where to place it within the menu structure
			raise NotDesktop("Categories entry is absent or invalid")

		try :
			self.terminal = _bool(x["Terminal"])
		except KeyError :
			self.terminal = False
			# Vast majority of the apps that have their .desktop files are the GUI ones
			# which aren't in need of console - that's why False is the default

		try :
			if _bool(x["NoDisplay"]) :
				raise NotDesktop("NoDisplay is true")
		except KeyError :
			pass

		try :
			if _bool(x["Hidden"]) :
				raise NotDesktop("Hidden is true")
		except KeyError :
			pass

		self.desktop = desktop # Source .desktop file

		# Some 3rd party packages
		if fnmatch.fnmatch(self.exename, "*kcmshell") :
			self.keywords |= KwS(KDE, Kw("X-KDE-settings"))

		super(Application, self).__init__()
